Package: hetGP
Type: Package
Title: Heteroskedastic Gaussian Process Modeling and Design under
        Replication
Version: 1.1.9
Date: 2026-04-17
Authors@R: c(person(given = "Mickael",
                        family = "Binois",
                        role = c("aut", "cre"),
                        email = "mickael.binois@inria.fr"),
                 person(given = c("Robert", "B."),
                        family = "Gramacy",
                        role = "aut"))
Description: Performs Gaussian process regression with heteroskedastic noise following the model by Binois, M., Gramacy, R., Ludkovski, M. (2016) <doi:10.48550/arXiv.1611.05902>, with implementation details in Binois, M. & Gramacy, R. B. (2021) <doi:10.18637/jss.v098.i13>. The input dependent noise is modeled as another Gaussian process. Replicated observations are encouraged as they yield computational savings. Sequential design procedures based on the integrated mean square prediction error and lookahead heuristics are provided, and notably fast update functions when adding new observations.
License: LGPL
LazyData: FALSE
Depends: R (>= 2.10),
Imports: Rcpp (>= 0.12.3), MASS, methods, DiceDesign, mco, quadprog
LinkingTo: Rcpp
Suggests: knitr, monomvn, lhs, colorspace
VignetteBuilder: knitr
RoxygenNote: 7.3.3
NeedsCompilation: yes
Packaged: 2026-04-23 15:02:57 UTC; mbinois
Author: Mickael Binois [aut, cre],
  Robert B. Gramacy [aut]
Maintainer: Mickael Binois <mickael.binois@inria.fr>
Repository: CRAN
Date/Publication: 2026-04-23 15:50:02 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-04-24 13:24:11 UTC; unix
Archs: hetGP.so.dSYM
