# Generated by roxygen2: do not edit by hand

S3method(plot,CRNGP)
S3method(plot,hetGP)
S3method(plot,hetTP)
S3method(plot,homGP)
S3method(plot,homTP)
S3method(predict,CRNGP)
S3method(predict,hetGP)
S3method(predict,hetTP)
S3method(predict,homGP)
S3method(predict,homTP)
S3method(print,CRNGP)
S3method(print,hetGP)
S3method(print,hetTP)
S3method(print,homGP)
S3method(print,homTP)
S3method(print,summary.CRNGP)
S3method(print,summary.hetGP)
S3method(print,summary.hetTP)
S3method(print,summary.homGP)
S3method(print,summary.homTP)
S3method(rebuild,hetGP)
S3method(rebuild,hetTP)
S3method(rebuild,homGP)
S3method(rebuild,homTP)
S3method(simul,CRNGP)
S3method(summary,CRNGP)
S3method(summary,hetGP)
S3method(summary,hetTP)
S3method(summary,homGP)
S3method(summary,homTP)
S3method(update,hetGP)
S3method(update,hetTP)
S3method(update,homGP)
S3method(update,homTP)
export(IMSPE)
export(IMSPE_optim)
export(LOO_preds)
export(Wij)
export(allocate_mult)
export(allocq)
export(allocq_c)
export(compareGP)
export(cov_gen)
export(crit_EI)
export(crit_ICU)
export(crit_IMSPE)
export(crit_MCU)
export(crit_MEE)
export(crit_cSUR)
export(crit_logEI)
export(crit_optim)
export(crit_qEI)
export(crit_tMSE)
export(deriv_crit_EI)
export(deriv_crit_IMSPE)
export(deriv_crit_logEI)
export(f1d)
export(f1d2)
export(f1d2_n)
export(f1d_n)
export(find_reps)
export(horizon)
export(hyperSharpeMax)
export(hyperSharperQ)
export(logLikH)
export(mleCRNGP)
export(mleHetGP)
export(mleHetTP)
export(mleHomGP)
export(mleHomTP)
export(pred_noisy_input)
export(qEI_loop)
export(qHSRI_loop)
export(rebuild)
export(scores)
export(simul)
export(sirEval)
export(sirSimulate)
export(strip)
export(update_pred)
import(methods)
importFrom(DiceDesign,lhsDesign)
importFrom(DiceDesign,maximinSA_LHS)
importFrom(MASS,ginv)
importFrom(Rcpp,evalCpp)
importFrom(graphics,abline)
importFrom(graphics,arrows)
importFrom(graphics,axis)
importFrom(graphics,filled.contour)
importFrom(graphics,legend)
importFrom(graphics,pairs)
importFrom(graphics,plot)
importFrom(graphics,points)
importFrom(mco,nsga2)
importFrom(methods,is)
importFrom(parallel,mclapply)
importFrom(quadprog,solve.QP)
importFrom(stats,cov2cor)
importFrom(stats,dnorm)
importFrom(stats,dt)
importFrom(stats,optim)
importFrom(stats,pnorm)
importFrom(stats,pt)
importFrom(stats,qnorm)
importFrom(stats,qt)
importFrom(stats,quantile)
importFrom(stats,rnorm)
importFrom(stats,runif)
importFrom(stats,uniroot)
importFrom(stats,var)
useDynLib(hetGP, .registration = TRUE)
