% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/IMSE.R
\name{deriv_crit_IMSPE}
\alias{deriv_crit_IMSPE}
\title{Derivative of crit_IMSPE}
\usage{
deriv_crit_IMSPE(x, model, Wijs = NULL)
}
\arguments{
\item{x}{matrix for the new design (size 1 x d)}

\item{model}{\code{homGP} or \code{hetGP} model}

\item{Wijs}{optional previously computed matrix of Wijs, see \code{\link[hetGP]{Wij}}}
}
\value{
Derivative of the sequential IMSPE with respect to \code{x}
}
\description{
Derivative of crit_IMSPE
}
\seealso{
\code{\link[hetGP]{crit_IMSPE}} for the criterion
}
